#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>

#include "data.h"

#define DISP_DATA _BV(PB0)
#define DISP_SHIFT _BV(PB1)
#define DISP_LATCH _BV(PB2)

unsigned char displayState[4];

void setDisplayInt(int num) {
	unsigned char i, digit, first = 1, negative = 0;
	unsigned int m;

	if(num < 0) {
		num = -num;
		negative = 1;
	}
	if(num > 9999) num = 9999;

	for(i = 0; i < 4; i++) {
		displayState[3-i] = pgm_read_byte(&segData[17]); /* blank */
		m = pgm_read_word(&mags[i]);
		for(digit = 0; num >= m; num -= m) digit++;
		if(first && digit < 1 && (3-i) > 1) continue;
		first = 0;
		displayState[3-i] = pgm_read_byte(&segData[digit]);
	}

	if(negative) {
		displayState[3] = pgm_read_byte(&segData[14]);
	}
}
/*
void setDisplayHex(unsigned int n) {
	displayState[0] = pgm_read_byte(&segData[n & 0x0f]);
	displayState[1] = pgm_read_byte(&segData[(n >> 4) & 0x0f]);
	displayState[2] = pgm_read_byte(&segData[(n >> 8) & 0x0f]);
	displayState[3] = pgm_read_byte(&segData[(n >> 12) & 0x0f]);
}
*/
void updateDisplay() {
	unsigned char i, j;
	
	for(i = 0; i < 4; i++) {
		for(j = 0; j < 4; j++) {
			if(i == (3 - j)) PORTB |= DISP_DATA;
			else PORTB &= (unsigned char)~DISP_DATA;
			PORTB |= DISP_SHIFT;
			PORTB &= (unsigned char)~DISP_SHIFT;
		}
		for(j = 0; j < 8; j++) {
			if((displayState[i] >> (7 - j)) & 0x1) PORTB |= DISP_DATA;
			else PORTB &= (unsigned char)~DISP_DATA;
			PORTB |= DISP_SHIFT;
			PORTB &= (unsigned char)~DISP_SHIFT;
		}

		PORTB |= DISP_LATCH;
		PORTB &= (unsigned char)~DISP_LATCH;
	}
}

unsigned int zeroPoint = 0;

unsigned int getPower() {
	unsigned long volts;

	volts = ADCL;
	volts |= ADCH << 8;
	volts *= 48876;
	volts /= 10000;

	return volts * volts / 10000;
}

ISR(TIM0_COMPA_vect) {
	unsigned long currentPower;

	currentPower = getPower();

	setDisplayInt((zeroPoint - currentPower) & 0xffff);
	displayState[1] |= pgm_read_byte(&segData[16]);
}

ISR(PCINT0_vect) {
	zeroPoint = getPower();
}

void init() {
	TCCR0A = _BV(WGM01);		// Clear timer on compare match mode
	TCCR0B = _BV(CS02) | _BV(CS00);	// Clock prescale by 1024
	OCR0A = 255;			// Compare match at top
	TIMSK0 = _BV(OCIE0A);		// Enable compare match interrupts

	ADMUX = _BV(MUX1);			// ADC2/PB4 as input, LSB aligned, Vcc ref
	ADCSRA = _BV(ADEN) | _BV(ADPS2);	// enable ADC with and /16 prescale
	ADCSRB = 0;				// free-run mode
	ADCSRA |= _BV(ADATE) | _BV(ADSC);	// start converting

	PCMSK |= _BV(PCINT3);		// interrupt on PB3 change
	GIMSK |= _BV(PCIE);		// pin change interrupts on
	PORTB |= _BV(PB3);  		// input pull-up

	DDRB = DISP_DATA | DISP_SHIFT | DISP_LATCH;	// display drive outputs

	sei();
}


int main() {
	init();

	while(1) updateDisplay();

	return 0;  // unreached
}
